/* PlanetPlanetPlanetPlanetPlanet is a blog aggregating program that uses too
   many C++ libraries.
   Copyright (C) 2006 James A. Morrison  <jim.morrison@gmail.com>

PlanetPlanetPlanetPlanetPlanet is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or
(at your option) any later version.

PlanetPlanetPlanetPlanetPlanet is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.  */

#include <error.h>
#include <fcntl.h>

#include <iostream>
#include <map>
#include <string>
#include <vector>

#include <expat.h>

#include <google/template.h>
#include <google/template_dictionary.h>

#include <wvstreams/uniconfroot.h>
#include <wvstreams/wvargs.h>
#include <wvstreams/wvfile.h>
#include <wvstreams/wvhttp.h>
#include <wvstreams/wvistreamlist.h>
#include <wvstreams/wvlog.h>
#include <wvstreams/wvlogrcv.h>
#include <wvstreams/wvpipe.h>
#include <wvstreams/wvstring.h>
#include <wvstreams/wvstrutils.h>

RegisterTemplateFilename(MAIN_PAGE_FN, "output/main_page.tpl");
#include "genfiles/main_page.tpl.varnames.h"

RegisterTemplateFilename(SINGLE_RESULT_FN, "output/single_result.tpl");
#include "genfiles/single_result.tpl.varnames.h"

using namespace google;

namespace {

// This class stores all the information to about one blog post.
class BlogPost {
 public:
  BlogPost(const char* username) : username_(username) { }

  const char* username() const { return username_; }

  const std::string& title() const { return title_; }
  void set_title(const char* title) { title_ = title; }
  void append_title(const char* title, int length) {
    title_.append(title, length);
  }

  const std::string& href() const { return href_; }
  void set_href(const char* href) { href_ = href; }
  void append_href(const char* href, int length) { href_.append(href, length); }

  const std::string& content() const { return content_; }
  void clear_content() { content_.clear(); }
  void add_content(const char* content, int length) {
    content_.append(content, length);
  }

  const std::string& time() const { return time_; }
  void append_time(const char* content, int len) { time_.append(content, len); }

  void clear();

 private:
  const char* const username_;
  std::string content_;
  std::string href_;
  std::string time_;
  std::string title_;
};

void BlogPost::clear() {
  content_.clear();
  href_.clear();
  title_.clear();
  time_.clear();
}

// This is the meat of PlanetPlanetPlanetPlanetPlanet.  The Blogs class
// reads data from a WvStream, streams it through Expat, keeps the state of
// parsing for a given feed and creates a BlogPost object for each post in
// the feed.  This class also saves the into into a cache file.

// This code is not thread safe, but we don't need threads with WvStreams,
// we've got coroutines.
class Blogs {
 public:
  Blogs(std::map<time_t, BlogPost>* bloglist, WvStringParm username,
        WvStringParm cache_filename);

  ~Blogs();

  void AddPost(WvStream* conn);
  void StartTag(const XML_Char* tag, const XML_Char** attrs);
  void EndTag(const XML_Char* tag);
  void AddContent(const XML_Char* content, int len);

  const std::string& feed_href() const { return feed_href_; }
  const std::string& feed_name() const { return feed_name_; }
  const WvString& username() const { return username_; }

 private:
  // Shared map
  std::map<time_t, BlogPost>* bloglist_;

  enum State {
    WAITING = 0,
    OUTER_FEED,
    IN_FEED_LINK,

    // Atom parsing states.
    IN_ENTRY,
    IN_ISSUED,
    IN_CONTENT,

    // RSS (livejournal) parsing states
    IN_ITEM,
    IN_DATE,
    IN_TITLE,
    IN_LINK,
    IN_DESCRIPTION,
  };

  void publish_blog();
  void set_date(const char* format);

  WvString username_;
  std::string feed_href_;
  std::string feed_name_;
  XML_Parser parser_;
  State state_;
  WvLog log_;
  WvFile cache_;

  void set_state(State state) {
    log_(WvLog::Debug5, "State transition %s -> %s\n", state_, state);
    state_ = state;
  }

  BlogPost current_post_;

  time_t post_time_;
};

static void xml_start_callback(void* userdata, const XML_Char* tag,
                               const XML_Char **attributes) {
  Blogs* blogs = (Blogs*)userdata;
  blogs->StartTag(tag, attributes);
}

static void xml_end_callback(void* userdata, const XML_Char* tag) {
  Blogs* blogs = (Blogs*)userdata;
  blogs->EndTag(tag);
}

static void content_callback(void* userdata, const XML_Char* content, int len) {
  Blogs* blogs = (Blogs*)userdata;
  blogs->AddContent(content, len);
}

Blogs::Blogs(std::map<time_t, BlogPost>* bloglist, WvStringParm username,
             WvStringParm cache_filename) 
  : bloglist_(bloglist),
    username_(username),
    state_(WAITING),
    log_("Blogs", WvLog::Info),
    cache_(cache_filename,  O_WRONLY | O_CREAT),
    current_post_(username) {
  parser_ = XML_ParserCreate(NULL);

  XML_SetUserData(parser_, this);
  XML_SetElementHandler(parser_, xml_start_callback, xml_end_callback);
  XML_SetCharacterDataHandler(parser_, content_callback);

}

Blogs::~Blogs() {
  XML_ParserFree(parser_);
}

void Blogs::AddPost(WvStream* conn) {
  void* expat_buffer = XML_GetBuffer(parser_, 8192);
  size_t read = conn->read(expat_buffer, 8192);
  if (state_ == WAITING && read > 0) {
    set_state(OUTER_FEED);
  }

  if (state_ != WAITING && read >= 0) {
    cache_.write(expat_buffer, read);
    XML_ParseBuffer(parser_, read, read == 0);
  }
}

static bool get_link_title_and_href(const XML_Char** attrs,
                                    const char** out_title,
                                    const char** out_href) {
  const char* title = "";
  const char* href = "";
  bool use_entry = false;
  for (int i = 0; attrs[i]; i += 2) {
    if (strcasecmp("title", attrs[i]) == 0) {
      title = attrs[i + 1];
    } else if (strcasecmp("href", attrs[i]) == 0) {
      href = attrs[i + 1];
    } else if (strcasecmp("rel", attrs[i]) == 0 &&
      strcasecmp("alternate", attrs[i + 1]) == 0) {
      use_entry = true;
    }
  }
  *out_title = title;
  *out_href = href;
  return use_entry;
}

void Blogs::StartTag(const XML_Char* tag, const XML_Char** attrs) {
  switch (state_) {
  case OUTER_FEED:
    if (strcasecmp("entry", tag) == 0) {
      set_state(IN_ENTRY);
    } else if (strcasecmp("item", tag) == 0) {
      set_state(IN_ITEM);
    } else if (strcasecmp("link", tag) == 0) {
      feed_href_.clear();
      feed_name_.clear();

      const char* title = "";
      const char* href = "";
      if (get_link_title_and_href(attrs, &title, &href)) {
        feed_name_ = title;
        feed_href_ = href;
      }

      set_state(IN_FEED_LINK);
    }
    break;

  case IN_ENTRY:
    if (strcasecmp("link", tag) == 0) {
      const char* title = "";
      const char* href = "";
      if (get_link_title_and_href(attrs, &title, &href)) {
        current_post_.set_title(title);
        current_post_.set_href(href);
      }
    } else if (strcasecmp("content", tag) == 0) {
      set_state(IN_CONTENT);
    } else if (strcasecmp("issued", tag) == 0) {
      set_state(IN_ISSUED);
    }
    break;

  case IN_ITEM:
    if (strcasecmp("link", tag) == 0) {
      set_state(IN_LINK);
    } else if (strcasecmp("title", tag) == 0) {
      set_state(IN_TITLE);
    } else if (strcasecmp("description", tag) == 0) {
      set_state(IN_DESCRIPTION);
    } else if (strcasecmp("content:encoded", tag) == 0) {
      current_post_.clear_content();
      set_state(IN_DESCRIPTION);
    } else if (strcasecmp("pubdate", tag) == 0) {
      set_state(IN_DATE);
    }
    break;
  }

}

void Blogs::EndTag(const XML_Char* tag) {
  switch (state_) {
  // RSS transitions.
  case IN_ITEM:
    if (strcasecmp("item", tag) == 0) { publish_blog(); }
    break;
  case IN_DESCRIPTION:
    if (strcasecmp("description", tag) == 0) { set_state(IN_ITEM); }
    else if (strcasecmp("content:encoded", tag) == 0) { set_state(IN_ITEM); }
    break;
  case IN_LINK:
    if (strcasecmp("link", tag) == 0) { set_state(IN_ITEM); }
    break;
  case IN_TITLE:
    if (strcasecmp("title", tag) == 0) { set_state(IN_ITEM); }
    break;
  case IN_DATE:
    if (strcasecmp("pubdate", tag) == 0) {
      set_date("%a, %d %b %Y %T %Z");
      set_state(IN_ITEM);
    }
    break;

  // Atom transitions.
  case IN_CONTENT:
    if (strcasecmp("content", tag) == 0) { set_state(IN_ENTRY); }
    break;
  case IN_ENTRY:
    if (strcasecmp("entry", tag) == 0) { publish_blog(); }
    break;
  case IN_ISSUED:
    if (strcasecmp("issued", tag) == 0) {
      set_date("%FT%T%z");
      set_state(IN_ENTRY);
    }
    break;

  // Generic transitions.
  case IN_FEED_LINK:
    if (strcasecmp("link", tag) == 0) { set_state(OUTER_FEED); }
    break;
  }

}

void Blogs::AddContent(const XML_Char* content, int len) {
  switch (state_) {
  case IN_CONTENT:
  case IN_DESCRIPTION:
    current_post_.add_content(content, len);
    break;
  case IN_ISSUED:
  case IN_DATE:
    current_post_.append_time(content, len);
    break;
  case IN_LINK:
    current_post_.append_href(content, len);
    break;
  case IN_TITLE:
    current_post_.append_title(content, len);
    break;
  case IN_FEED_LINK:
    feed_href_.append(content, len);
    break;
  }
}

// convert time_str into a time_t
void Blogs::set_date(const char* format) {
  struct tm now;
  if (!strptime(current_post_.time().c_str(), format, &now)) {
    log_(WvLog::Warning, "Unable to parse \"%s\" as time string\n",
         current_post_.time().c_str());
  }

  post_time_ = mktime(&now);
}

void Blogs::publish_blog() {
  if (bloglist_->find(post_time_) != bloglist_->end()) {
    log_(WvLog::Warning,  "two posts in the same second\n");
  }

  bloglist_->insert(std::make_pair(post_time_, current_post_));
  current_post_.clear();
  set_state(OUTER_FEED);
}

}  // anonymous namespace

static std::string FillResults(const std::map<time_t, BlogPost>& entries,
                               const std::vector<Blogs*>& blogs,
                               const UniConf& config,
                               bool strip_whitespace) {
  int displayed = 0;
  int max_results = config.xgetint("maximum posts", entries.size());

  WvLog log("FillResults", WvLog::Debug4);
  TemplateDictionary dictionary("Outer dictionary");
  dictionary.SetValue(kmp_GENERATOR, "PlanetPlanetPlanetPlanetPlanet");

  static const int columns = config.xgetint("subscription columns", 6);
  WvString name = config.xget("name");
  dictionary.SetValue(kmp_PLANET_NAME, name.cstr());

  int subscriptions = 0;
  TemplateDictionary *row_dict;
  for (std::vector<Blogs*>::const_iterator i = blogs.begin();
       i != blogs.end(); ++i, ++subscriptions) {
    if (subscriptions % columns == 0) {
      row_dict = dictionary.AddSectionDictionary(kmp_SUBSCRIPTIONS_ROW);
    }

    TemplateDictionary* dict =
        row_dict->AddSectionDictionary(kmp_SUBSCRIPTION_CELL);
    dict->SetValueAndShowSection(kmp_FEED_URL, (*i)->feed_href(),
                                 kmp_SUBSCRIPTION_LINK);
    dict->SetValue(kmp_USERNAME, (*i)->username().cstr());
    dict->SetValue(kmp_FEED_TITLE, (*i)->feed_name());
  }

  dictionary.SetIntValue(kmp_COLUMNS,
                         subscriptions < columns ? subscriptions : columns);

  for (std::map<time_t, BlogPost>::const_reverse_iterator i = entries.rbegin();
       i != entries.rend() && displayed < max_results; ++i, ++displayed) {
    TemplateDictionary* result =
      dictionary.AddIncludeDictionary(kmp_SINGLE_POST);
    log("%s\n", i->second.title().c_str());
    result->SetFilename(SINGLE_RESULT_FN);
    result->SetValue(ksr_BLOGGER_NAME, i->second.username());
    result->SetValue(ksr_POST_TITLE, i->second.title());
    result->SetValue(ksr_POST_LINK, i->second.href());
    result->SetValue(ksr_POST_TIME, i->second.time());
    result->SetValue(ksr_BLOG_CONTENT, i->second.content());
  }

  Template* tpl =
      Template::GetTemplate(MAIN_PAGE_FN,
                            strip_whitespace ? STRIP_WHITESPACE : DO_NOT_STRIP);
  if (!tpl) {
    return "";
  }

  std::string output_buffer;
  tpl->Expand(&output_buffer, &dictionary);
  return output_buffer;
}

static void read_callback(WvStream& conn, void* userdata) {
  if (!conn.isok()) {
    return;
  }

  Blogs* blogs = (Blogs*)userdata;
  blogs->AddPost(&conn);
}

static bool decrease_logging(void* userdata)
{
  WvLog::LogLevel* level = static_cast<WvLog::LogLevel*>(userdata);
  if ((int)*level > (int)WvLog::Critical)
    *level = (WvLog::LogLevel)((int)*level - 1);
  return true;
}


static bool increase_logging(void* userdata)
{
  WvLog::LogLevel* level = static_cast<WvLog::LogLevel*>(userdata);
  if ((int)*level < (int)WvLog::Debug5)
    *level = (WvLog::LogLevel)((int)*level + 1);
  return true;
}

// Do a HEAD request for URL to see if it has changed.
// Write the the last_modified_filename if the feed has changed.
static bool unmodified(const WvUrl& url, WvStringParm last_modified_filename) {
  WvLog log("Unmodified", WvLog::Debug2);
  WvString last_modified_stamp;

  {
    WvFile last_modified_file(last_modified_filename, O_RDONLY);
    last_modified_stamp = trim_string(last_modified_file.getline());
    if (!last_modified_stamp) {
      log("No last modified stamp\n");
    }

    // The file may have been closed on us, so we are best to simply close
    // out and deal with writing to the file in a new stream.
    last_modified_file.close();
  }

  signal(SIGPIPE, SIG_IGN);

  char port[10];
  snprintf(port, sizeof(port), "%d", url.getport());
  const char* ncargs[] = { "-v", url.gethost().cstr(), port, 0 };
  WvPipe nc("/bin/nc", ncargs, true, true, false);
  nc.write(WvString("HEAD %s HTTP/1.0\r\nHost: %s\r\n\r\n",
                    url.getfile(), url.gethost()));

  if (!nc.isok()) {
    log(nc.errstr());
    return false;
  }


  // Assume if we got a non-HTTP response or any error response code that
  // the feed hasn't changed.
  char* line = trim_string(nc.getline(-1));
  if (strncmp(line, "HTTP/", 5) != 0) {
    log("Not an HTTP response: %s\n", line);
    return !!last_modified_stamp;
  }

  int response_code = atoi(trim_string(line+9));
  if (response_code >= 400) {
    log("Strange response code %s\n", response_code);
    return !!last_modified_stamp;
  }

  WvHTTPHeaderDict headers(15);
  while (line = nc.getline(-1)) {
    log(WvLog::Debug4, line);
    char *cptr = strchr(line, ':');
    if (cptr) {
      *cptr++ = 0;
      line = trim_string(line);
      cptr = trim_string(cptr);
      headers.add(new WvHTTPHeader(line, cptr), true);
    }
  }

  WvHTTPHeader* last_modified = headers["Last-Modified"];
  if (last_modified) {
    log("Last modified stamp %s\n", last_modified_stamp);
    log("Last modified header %s\n", last_modified->value);
    if (last_modified_stamp == last_modified->value) {
      // Don't bother changing the last modified file, it won't change.
      return true;
    } else {
      // Update the last modified file.
      WvFile last_modified_file(last_modified_filename, O_WRONLY | O_CREAT);
      last_modified_file.write(last_modified->value);
      last_modified_file.flush_then_close(-1);
    }
  }

  return false;
}

static void create_connection(const UniConf& config, const UniConf& sect,
                              std::vector<Blogs*>* blogs,
                              std::map<time_t, BlogPost>* blogposts) {
  // Create a blogs object for each file downloading
  //   A parser in each blogs object
  WvStream* conn = NULL;
  WvString uri = sect.xget("url");
  WvString file = sect.xget("file");
  WvString cache_file("%s/%s", config.xget("cache directory"), sect.key());
  WvLog log("Connection", WvLog::Info);
  if (!!file) {
    log("File:   ");
    conn = new WvFile(file, O_RDONLY);
  } else if (!!uri) {
    WvUrl url(uri);
    if (unmodified(url, WvString("%s.last_modified", cache_file))) {
      log("Cached: ");
      conn = new WvFile(cache_file, O_RDONLY);
    } else {
      log("URL:    ");
      WvHTTPStream* http = new WvHTTPStream(url);
      http->client_headers.add(
          new WvHTTPHeader(
              "User-Agent",
               WvString("http://www.csclub.uwaterloo.ca/"
                        "~ja2morri/bzr/ppppplanet/; %s",
                        config.xget("owner email"))), true);
      conn = http;
    }
  }

  if (conn) {
    log("Getting feed for %s\n", sect.key());
    blogs->push_back(new Blogs(blogposts, sect.key(), cache_file));
    conn->setcallback(read_callback, blogs->at(blogs->size() - 1));
    WvIStreamList::globallist.append(conn, true, "client");
  }
}

int main(int argc, char* argv[]) {
  WvArgs args;
  WvString config_file;
  WvStringList show_blogs; 
  WvLog::LogLevel log_level = WvLog::Info;
  bool strip_whitespace = false;

  args.add_option('c', "config", "Configuration to use", "ini:config_file",
                  config_file);
  args.add_option('b', "blog", "Show blog", "\"Jim Morrison\"", show_blogs);
  args.add_option('d', "debug", "Increase the amount of debug output",
                  increase_logging, &log_level);
  args.add_option('q', "quiet", "Decrease the amount of output",
                  decrease_logging, &log_level);
  args.add_set_bool_option('s', "strip-whitespace", "Strip whitespace",
                           strip_whitespace);
  args.process(argc, argv);

  // Postprocess the arguments.
  if (!config_file) {
    // attempting to use $HOME/.planetplanetplanetplanetplanet
    config_file = "ini:";
    config_file.append(getenv("HOME"));
    config_file.append("/.planetplanetplanetplanetplanet");
  }

  // Load a config file (uniconf)
  UniConfRoot cfg(config_file);
  if (!cfg.haschildren()) {
    error(1, errno, "Cannot load data from %s", config_file.cstr());
    return 1;
  }

  WvLogConsole log_receiver(1, log_level);
  UniConf config = cfg["Config"];

  std::map<time_t, BlogPost> blogposts;
  std::vector<Blogs*> blogs;
  WvIStreamList conns;

  if (show_blogs.isempty()) {
    UniConf::Iter i(cfg.root());
    for (i.rewind();i.next();) {
      create_connection(config, *i.ptr(), &blogs, &blogposts);
    }
  } else {
    WvStringList::Iter i(show_blogs);
    for (i.rewind(); i.next();) {
      create_connection(config, cfg[*i.ptr()], &blogs, &blogposts);
    }
  }

  // go into the WvStreams select loop
  while (!WvIStreamList::globallist.isempty())
    WvIStreamList::globallist.runonce(1000);

  std::string html_buffer =
      FillResults(blogposts, blogs, config,
                  strip_whitespace);

  for (int i = 0; i < blogs.size(); ++i) {
    delete blogs[i];
  }

  WvString output = config.xget("output file");
  if (!!output) {
    FILE* f = fopen(output.cstr(), "w");
    if (!f) {
      error(1, errno, "%s", output.cstr());
    }

    fwrite(html_buffer.c_str(), html_buffer.size(), 1, f);
    fclose(f);
  } else {
    // Write html_buffer somewhere.
    std::cout << html_buffer;
    std::cout << std::endl;
  }

  return 0;
}
